use std::{collections::HashMap, rc::Rc, sync::Mutex};

use crate::{
    lexer::Lexer,
    parsed::{Node, NodeKind},
    pos::{Error, ErrorAcc, Pos},
    token::{Token, TokenKind, TokenValue},
};

pub struct Parser<'a> {
    lexer: Lexer<'a>,
    current: Option<Token>,
    error_acc: Rc<Mutex<ErrorAcc>>,
}

impl<'a> Parser<'a> {
    pub fn new(text: &'a str, error_acc: Rc<Mutex<ErrorAcc>>) -> Self {
        let mut lexer = Lexer::new(text, error_acc.clone());
        let current = lexer.next();
        Self {
            lexer,
            current,
            error_acc,
        }
    }

    pub fn parse(&mut self) -> Vec<Node> {
        self.parse_file()
    }

    pub fn symbols(self) -> HashMap<u64, String> {
        self.lexer.symbols()
    }

    fn parse_file(&mut self) -> Vec<Node> {
        let mut stmts = Vec::new();
        loop {
            match self.current {
                Some(_) => stmts.push(self.parse_stmt()),
                None => break stmts,
            }
        }
    }

    fn parse_stmt(&mut self) -> Node {
        let pos = self.pos().unwrap();
        match self.curr_kind() {
            Some(TokenKind::LBrace) => self.parse_block(),
            Some(TokenKind::If) => self.parse_if(),
            Some(TokenKind::Loop) => self.parse_loop(),
            Some(TokenKind::Fn) => self.parse_fn(),
            _ => {
                let stmt = match self.curr_kind() {
                    Some(TokenKind::Let) => self.parse_let(),
                    Some(TokenKind::Break) => {
                        self.step();
                        self.node(NodeKind::Break, pos.clone())
                    }
                    Some(TokenKind::Return) => {
                        self.step();
                        let value = match self.curr_kind() {
                            Some(TokenKind::Semicolon) => None,
                            _ => Some(Box::new(self.parse_expr())),
                        };
                        self.node(NodeKind::Return { value }, pos.clone())
                    }
                    _ => self.parse_assign(),
                };
                match self.curr_kind() {
                    Some(TokenKind::Semicolon) => {
                        self.step();
                        stmt
                    }
                    _ => {
                        self.error("expected ';'", pos.clone());
                        self.error_node(pos)
                    }
                }
            }
        }
    }

    fn parse_fn(&mut self) -> Node {
        let pos = self.pos().unwrap();
        self.step();
        if !self.curr_is(TokenKind::Id) {
            self.error("expected id", pos.clone());
            return self.error_node(pos);
        }
        let subject = Box::new(self.parse_id());
        if !self.curr_is(TokenKind::LParen) {
            self.error("expected '('", pos.clone());
            return self.error_node(pos);
        }
        let params = match self.parse_fn_params() {
            Ok(params) => params,
            Err(expr) => return expr,
        };
        if !self.curr_is(TokenKind::MinusLt) {
            self.error("expected '->'", pos.clone());
            return self.error_node(pos);
        }
        self.step();
        let return_typ = Box::new(self.parse_typ());
        if !self.curr_is(TokenKind::LBrace) {
            self.error("expected '{'", pos.clone());
            return self.error_node(pos);
        }
        let body = Box::new(self.parse_block());
        self.node(
            NodeKind::Fn {
                subject,
                params,
                return_typ,
                body,
            },
            pos,
        )
    }

    fn parse_fn_params(&mut self) -> Result<Vec<Node>, Node> {
        let pos = self.pos().unwrap();
        self.step();
        let mut params = Vec::new();
        if !self.curr_is(TokenKind::RParen) {
            if !self.curr_is(TokenKind::Id) {
                self.error("expected id", pos.clone());
                return Err(self.error_node(pos));
            }
            params.push(self.parse_param());
            while let Some(TokenKind::Comma) = self.curr_kind() {
                self.step();
                if self.curr_is(TokenKind::RParen) {
                    self.error("expected ')'", pos.clone());
                    break;
                }
                params.push(self.parse_param());
            }
        }
        if !self.curr_is(TokenKind::RParen) {
            self.error("expected ')'", pos.clone());
            return Err(self.error_node(pos));
        }
        self.step();
        Ok(params)
    }

    fn parse_let(&mut self) -> Node {
        let pos = self.pos().unwrap();
        self.step();
        if !self.curr_is(TokenKind::Id) {
            self.error("expected id", pos.clone());
            return self.error_node(pos);
        }
        let subject = self.parse_param();
        if !self.curr_is(TokenKind::Equal) {
            self.error("expected '='", pos.clone());
            return self.error_node(pos);
        }
        self.step();
        let value = self.parse_expr();
        self.node(
            NodeKind::Let {
                subject: Box::new(subject),
                value: Box::new(value),
            },
            pos,
        )
    }

    fn parse_param(&mut self) -> Node {
        let pos = self.pos().unwrap();
        let subject = Box::new(self.parse_id());
        let typ = if let Some(TokenKind::Colon) = self.curr_kind() {
            self.step();
            Some(Box::new(self.parse_typ()))
        } else {
            None
        };
        self.node(NodeKind::Param { subject, typ }, pos)
    }

    fn parse_typ(&mut self) -> Node {
        let pos = self.pos().unwrap();
        match self.curr_kind() {
            Some(TokenKind::Id) => self.parse_id(),
            _ => {
                self.error("expected type", pos.clone());
                self.step();
                self.error_node(pos)
            }
        }
    }

    fn parse_assign(&mut self) -> Node {
        let pos = self.pos().unwrap();
        let subject = self.parse_expr();
        match self.curr_kind() {
            Some(TokenKind::Equal) => {
                self.step();
                let value = self.parse_expr();
                self.node(
                    NodeKind::Assign {
                        subject: Box::new(subject),
                        value: Box::new(value),
                    },
                    pos,
                )
            }
            _ => subject,
        }
    }

    fn parse_expr(&mut self) -> Node {
        self.parse_call()
    }

    fn parse_call(&mut self) -> Node {
        let pos = self.pos().unwrap();
        let mut subject = self.parse_value();
        loop {
            match self.curr_kind() {
                Some(TokenKind::LParen) => {
                    self.step();
                    let mut args = Vec::new();
                    match self.curr_kind() {
                        None | Some(TokenKind::RParen) => {}
                        Some(_) => {
                            args.push(self.parse_expr());
                            while let Some(TokenKind::Comma) = self.curr_kind() {
                                self.step();
                                if let Some(TokenKind::RParen) = self.curr_kind() {
                                    break;
                                }
                                args.push(self.parse_expr());
                            }
                        }
                    }
                    match self.curr_kind() {
                        Some(TokenKind::RParen) => {}
                        _ => {
                            self.error("expected ')'", pos.clone());
                            return self.error_node(pos);
                        }
                    }
                    self.step();
                    subject = self.node(
                        NodeKind::Call {
                            subject: Box::new(subject),
                            args,
                        },
                        pos.clone(),
                    );
                }
                _ => break subject,
            }
        }
    }

    fn parse_value(&mut self) -> Node {
        let pos = self.pos().unwrap();
        match self.curr_kind() {
            Some(TokenKind::Id) => self.parse_id(),
            Some(TokenKind::Int) => self.parse_int(),
            Some(TokenKind::Str) => self.parse_string(),
            Some(TokenKind::LParen) => self.parse_group(),
            Some(TokenKind::LBrace) => self.parse_block(),
            Some(TokenKind::If) => self.parse_if(),
            Some(TokenKind::Loop) => self.parse_loop(),
            _ => {
                self.error("expected value", pos.clone());
                self.step();
                self.error_node(pos)
            }
        }
    }

    fn parse_id(&mut self) -> Node {
        let pos = self.pos().unwrap();
        let Some(Token {
            kind: TokenKind::Id,
            value: TokenValue::Id(value),
            ..
        }) = self.current
        else {
            unreachable!()
        };
        self.step();
        self.node(NodeKind::Id(value), pos)
    }

    fn parse_int(&mut self) -> Node {
        let pos = self.pos().unwrap();
        let Some(Token {
            kind: TokenKind::Int,
            value: TokenValue::Int(value),
            ..
        }) = self.current
        else {
            unreachable!()
        };
        self.step();
        self.node(NodeKind::Int(value), pos)
    }

    fn parse_string(&mut self) -> Node {
        let pos = self.pos().unwrap();
        let Some(Token {
            kind: TokenKind::Str,
            value: TokenValue::Str(value),
            ..
        }) = self.current.clone()
        else {
            unreachable!()
        };
        self.step();
        self.node(NodeKind::Str(value.clone()), pos)
    }

    fn parse_group(&mut self) -> Node {
        let pos = self.pos().unwrap();
        self.step();
        let expr = Box::new(self.parse_expr());
        if !self.curr_is(TokenKind::RParen) {
            self.error("expected ')'", pos.clone());
            return self.error_node(pos);
        }
        self.step();
        self.node(NodeKind::Group(expr), pos)
    }

    fn parse_block(&mut self) -> Node {
        let pos = self.pos().unwrap();
        self.step();
        let mut stmts = Vec::new();
        loop {
            match self.curr_kind() {
                None => {
                    self.error("expected ')'", pos.clone());
                    break self.error_node(pos);
                }
                Some(TokenKind::RBrace) => {
                    self.step();
                    break self.node(NodeKind::Block(stmts), pos);
                }
                _ => stmts.push(self.parse_stmt()),
            }
        }
    }

    fn parse_if(&mut self) -> Node {
        let pos = self.pos().unwrap();
        self.step();
        let cond = Box::new(self.parse_expr());
        if !self.curr_is(TokenKind::LBrace) {
            self.error("expected '}'", pos.clone());
            return self.error_node(pos);
        }
        let truthy = Box::new(self.parse_block());
        let falsy = match self.curr_kind() {
            Some(TokenKind::Else) => {
                self.step();
                if !self.curr_is(TokenKind::LBrace) {
                    self.error("expected '}'", pos.clone());
                    return self.error_node(pos);
                }
                Some(Box::new(self.parse_block()))
            }
            _ => None,
        };
        self.node(
            NodeKind::If {
                cond,
                truthy,
                falsy,
            },
            pos,
        )
    }

    fn parse_loop(&mut self) -> Node {
        let pos = self.pos().unwrap();
        self.step();
        if !self.curr_is(TokenKind::LBrace) {
            self.error("expected '}'", pos.clone());
            return self.error_node(pos);
        }
        let body = Box::new(self.parse_block());
        self.node(NodeKind::Loop { body }, pos)
    }

    fn error<S: Into<String>>(&mut self, msg: S, pos: Pos) {
        let msg = msg.into();
        self.error_acc.lock().unwrap().add(Error {
            kind: crate::pos::ErrorKind::ParserError,
            pos: Some(pos),
            msg,
        });
    }

    fn step(&mut self) {
        self.current = self.lexer.next();
    }

    fn node(&self, kind: NodeKind, pos: Pos) -> Node {
        Node { kind, pos }
    }

    fn error_node(&self, pos: Pos) -> Node {
        Node {
            kind: NodeKind::Error,
            pos,
        }
    }

    fn pos(&self) -> Option<Pos> {
        self.current.as_ref().map(|token| token.pos.clone())
    }

    fn curr_is(&self, kind: TokenKind) -> bool {
        self.curr_kind() == Some(kind)
    }

    fn curr_kind(&self) -> Option<TokenKind> {
        self.current.as_ref().map(|t| t.kind.clone())
    }
}
